//
//  PurpleSkypeAccount.h
//  Adium
//

#import <AdiumLibpurple/CBPurpleAccount.h>

#define KEY_SKYPE_SHOW_SKYPEOUT		@"Skype:Show SkypeOut Contacts"
#define KEY_SKYPE_SYNC_OFFLINE		@"Skype:Sync Offline Status"
#define KEY_SKYPE_CHECK_FOR_UPDATES	@"Skype:Check for Updates"
#define KEY_SKYPE_AUTOSTART			@"Skype:Auto Start"

@interface PurpleSkypeAccount : CBPurpleAccount {

}

@end
